# -*- coding: latin-1 -*-

"""

This a partially implemented imap4 library, built on top of the standard 
imaplib. It's meant to be used instead of Pythons imaplib, which is pretty 
terrible.

I have only implemented the methods I need. Later I might implement the full 
spec.

"""

# python
import imaplib, re
from time import gmtime, strftime, time

# local library
from parseNestedParens import parseNestedParens
import imapUTF7
## debug version
#from Products.mxmImapUTF7 import imapUTF7

# exceptions
class IMAP4Exception(Exception):
    def __init__(self, *args):
        Exception.__init__(self, *args)
class IMAP4Failure(IMAP4Exception):
    "Failure, cannot execute command"
class IMAP4Command(IMAP4Exception):
    "Command unknown or arguments invalid"


# Flags Message Attribute
SEEN = r'\Seen'
ANSWERED = r'\Answered'
FLAGGED = r'\Flagged'
DELETED = r'\Deleted'
DRAFT = r'\Draft'
RECENT = r'\Recent'


class Message:
    "Returns an email message"


# Name attributes
NOINFERIORS = r'\Noinferiors'
NOSELECT = r'\Noselect'
MARKED = r'\Marked'
UNMARKED = r'\Unmarked'


class Folder:

    "An imap folder"

    def __init__(self, name=u'', flags=(), delimiter='/'):
        self.name = name
        self.flags = flags
        self.delimiter = delimiter

    def path(self):
        "Returns path as list"
        return self.name.split(self.delimiter)

    def flagged_with(self, flags):
        "Returns true if folder has one of flags"
        for flag in flags:
            if flag in self.flags:
                return 1
        return 0

    def __str__(self):
        return '\n'.join([
            'self.name: ' + repr(self.name),
            'self.flags: ' + repr(self.flags),
            'self.delimiter: ' + repr(self.delimiter),
        ])
        
        
# The 4 legal connection states
NON_AUTHENTICATED = 'non-authenticated'
AUTHENTICATED = 'authenticated'
SELECTED = 'selected'
LOGOUT = 'logout and close connection'

# parsers for results

list_response = re.compile('\((.*?)\) "(.*?)" "(.*?)"')
uid_match = re.compile(r'UID (\d+)')
size_match = re.compile(r'RFC.ZISE (\d+)')

def parseUid(resp):
    """Convert IMAP4 UID response to int."""
    mo = uid_match.search(resp)
    if not mo:
        return 0
    return int(mo.groups()[0])


def parseSize(resp):
    """Convert IMAP4 UID response to int."""
    mo = size_match.search(resp)
    if not mo:
        return 0
    return int(mo.groups()[0])


#############################
# command results
OK = 'OK'
NO = 'NO'
BAD = 'BAD'
response_map = {
    NO : IMAP4Failure,
    BAD : IMAP4Command,
}
def result_check(result):
    """
    General function for checking the result of a command, and raising the 
    apropriate Exception
    """
    typ, res = result
    if typ != OK:
        if not res: res = 'Unknown error'
        raise response_map.get(typ, IMAP4Exception)(res)


def ids2str(ids):
    return ','.join([str(id) for id in ids])


class ImapClientConnection:


    def __init__(self, connection):
        self.connection = connection
        self.state = AUTHENTICATED

    def _change_state(self, new_state):
        "Changes the state of the client"
        self.state = new_state

    def setAuthData(self, server_uri, user, password):
        "Set authentication data"
        self.server_uri = server_uri
        self.user = user
        self.password = password

    def __hash__(self):
        return hash((self.server_uri, self.user, self.password))

    # IMAP4 commands


#    def command(self):
#        "General command method"


#    def login(user, password):
#        self._change_state(AUTHENTICATED)


    def list(self, reference='""', mailbox_name='*', uid=None):
        "Returns a list of folders"
        typ, data = self.connection.list(reference, mailbox_name)
        result_check((typ, data))
        # Parse result, and return folder objects
        folders = []
        parsedParens = parseNestedParens(''.join(data))
        for i in range(0, len(parsedParens), 3):
            flags, delimiter, name = parsedParens[i:i+3]
            path = [f for f in name.split(delimiter)]
            folder = Folder(name.decode('imap4-utf-7'), flags, delimiter)
            folders.append(folder)
        return folders


    def select(self, folder='INBOX', readonly=None):
        "Selects a folder"
        typ, n_messages = self.connection.select(folder.encode('imap4-utf-7'), readonly)
        result_check((typ, n_messages))
        return int(n_messages[0])


    def fetch(self, message_set, message_parts, useUID=0):
        "Mother of all fetches"
        if not message_set: # don't fetch if there is nothing to fetch
            return []
        if useUID:
            typ, data = self.connection.uid('fetch', message_set, message_parts)
        else:
            typ, data = self.connection.fetch(message_set, message_parts)
        result_check((typ, data))
        return data


    def fetchRFC822(self, messageIds, useUID=0):
        """
        Fetches RFC822 messages. Returns list of (uid, rfc822Message).
        uid==true -> messageIds 'Unique Identifier (UID)'
        uid==False -> messageIds are 'Message Sequence Number'
        """
        messageIds = ids2str(messageIds)
        results = self.fetch(messageIds, '(RFC822)', useUID)
        # iterate through the results and build list of (UID, message) tuples
        messages = []
        if results != [None]:
            for i in range(0, len(results), 2):
                ((envelopeStart, msg), envelopeEnd) = results[i:i+2]
                envelope = envelopeStart + envelopeEnd
                uid = int(uid_match.search(envelope).groups()[0])
                messages.append((uid, msg))
        return messages


    def fetchBody(self, messageIds, useUID=0):
        """
        Fetches the body/structure of the messages and returns them as nested 
        parens.
        """
        messageIds = ids2str(messageIds)
        results = self.fetch(messageIds, '(BODY)', useUID)
        bodies = []
        if results != [None]:
            for result in results:
                n_msg, parens = parseNestedParens(result)
                mapper = {}
                for i in range(0, len(parens), 2):
                    key, val = parens[i:i+2]
                    mapper[key.upper()] = val
                bodies.append((mapper['UID'], mapper['BODY']))
        return bodies


    def fetchPart(self, messageId, part, useUID=0):
        """
        Fetches the part of body.
        eg. ('TEXT',) or (1,'TEXT') or (1,1,'HEADERS')
        """
        command = '(BODY[%s])' % '.'.join(part)
        results = self.fetch(str(messageId), command, useUID)
#        bodies = []
#        if results != [None]:
#            for result in results:
#                n_msg, parens = parseNestedParens(result)
#                mapper = {}
#                for i in range(0, len(parens), 2):
#                    key, val = parens[i:i+2]
#                    mapper[key.upper()] = val
#                bodies.append((mapper['UID'], mapper['BODY']))
#        return bodies


###    def fetchBodyStructure(self, messageIds, useUID=0):
###        """
###        Fetches the body/structure of the messages and returns them as nested 
###        parens.
###        """
###        messageIds = ids2str(messageIds)
###        results = self.fetch(messageIds, '(BODYSTRUCTURE)', useUID)
###        bodies = []
###        if results != [None]:
###            for result in results:
###                print result
####                n_msg, parens = parseNestedParens(result)
####                mapper = {}
####                for i in range(0, len(parens), 2):
####                    key, val = parens[i:i+2]
####                    mapper[key.upper()] = val
####                bodies.append((mapper['UID'], mapper['BODY']))
####        return bodies

#    def _parseBody(self):

    def parsedBody(self):
        """
        Returns a parsed body structure as a dict.
        part number in a tuple as key.
        """

    def fetchRFC822Text(self, messageIds, useUID=0):
        """
        Fetches text only content of RFC822 messages.
        Returns list of (uid, rfc822Message).
        uid==true -> messageIds 'Unique Identifier (UID)'
        uid==False -> messageIds are 'Message Sequence Number'
        """
        messageIds = ids2str(messageIds)
        results = self.fetch(messageIds, '(RFC822.TEXT)', useUID)
        # iterate through the results and build list of (UID, message) tuples
        messages = []
        if results != [None]:
            for i in range(0, len(results), 2):
                ((envelopeStart, msg), envelopeEnd) = results[i:i+2]
                envelope = envelopeStart + envelopeEnd
                uid = int(uid_match.search(envelope).groups()[0])
                messages.append((uid, msg))
        return messages


    def fetchBodyPart(self, messageId, part, useUID=0):
        """
        Fetches part of RFC822 message. Returns list of (uid, rfc822Part).
        uid==true -> messageIds 'Unique Identifier (UID)'
        uid==False -> messageIds are 'Message Sequence Number'
        """
        command = '(BODY[%s])' % '.'.join(part)
        results = self.fetch(str(messageId), command, useUID)
        # iterate through the results and build list of (UID, message) tuples
        try:
            messages = []
            if results != [None]:
                for i in range(0, len(results), 2):
                    ((envelopeStart, msg), envelopeEnd) = results[i:i+2]
                    envelope = envelopeStart + envelopeEnd
                    uid = int(uid_match.search(envelope).groups()[0])
                    messages.append((uid, msg))
            return messages[0][1]
        except:
            raise IMAP4Exception, 'No such body part'


    def fetchHeaders(self, messageIds, fields=None, useUID=0):
        """
        Returns the message headers
        """
        messageIds = ids2str(messageIds)
        if fields: # fields must be sequence of field names
            fields = ' '.join([field.upper() for field in fields])
            message_parts = '(BODY[HEADER.FIELDS (%s)])' % fields
        else:
            message_parts = '(BODY[HEADER])'
        response = self.fetch(messageIds, message_parts, useUID)
        results = []
        for i in range(0, len(response), 2):
            ((envelopeStart, headers), envelopeEnd) = response[i:i+2]
            envelope = envelopeStart + envelopeEnd
            uid = int(uid_match.search(envelope).groups()[0])
            results.append((uid, headers))
        return results


    def search(self, criteria=None, charset=None, useUID=0):
        """
        Search for messages with criteria
        msgnums = M.search(('FROM', '"LDJ"'))
        msgnums = M.search(('(FROM "LDJ")',))
        msguids = M.search(('(FROM "LDJ")',), useUID=1)
        """
        if not criteria: # don't fetch if there is nothing to fetch
            return []
        if useUID:
            typ, data = self.connection.uid('search', charset, *criteria)
        else:
            typ, data = self.connection.search(charset, *criteria)
        result_check((typ, data))
        return map(int, parseNestedParens(data[0]))


    def getAllUids(self):
        "Returns all uids in the selected folder"
        results = self.fetch('1:*', '(UID FLAGS RFC822.SIZE)')
        if results == [None]: # hmm! odd result
            results = []
        # iterate through the results and build list of (uid, flags, size) tuples
        infos = []
        for itm in results:
            uid = parseUid(itm)
            flags = imaplib.ParseFlags(itm)
            size = parseSize(itm)
            info = (uid, flags, size)
            infos.append(info)
        return infos


    def getMinimumUids(self, minimum, useUID=0):
        "Returns all uids in the selected folder with value larger than and equal to minimum"
        results = self.fetch('%s:*' % minimum, '(UID FLAGS RFC822.SIZE)', useUID)
        if results == [None]: # hmm! odd result
            results = []
        # iterate through the results and build list of (uid, flags, size) tuples
        infos = []
        for itm in results:
            uid = parseUid(itm)
            flags = imaplib.ParseFlags(itm)
            size = parseSize(itm)
            info = (uid, flags, size)
            infos.append(info)
        return infos

    ##################

    def store(self, message_set, command, flag_list, silent=0, useUID=0):
        "Set flags on the messages"
        message_set = ids2str(message_set)
        if silent: 
            command = command + '.SILENT'
        # convert flag list to imap bastard flag list/string
        flag_list = '(%s)' % ' '.join(flag_list)
        if useUID:
            typ, data = self.connection.uid('store', message_set, command, flag_list)
        else:
            typ, data = self.connection.store(message_set, command, flag_list)
        result_check((typ, data))
        return data


    def storeFlags(self, message_set, flag_list, silent=0, useUID=0):
        "Sets the flags on the message"
        self.store(message_set, 'FLAGS', flag_list, silent, useUID)
        

    def storeAddFlags(self, message_set, flag_list, silent=0, useUID=0):
        "Add flags to the message"
        return self.store(message_set, '+FLAGS', flag_list, silent, useUID)
        

    def storeSubtractFlags(self, message_set, flag_list, silent=0, useUID=0):
        "Subtract flags from the message"
        return self.store(message_set, '-FLAGS', flag_list, silent, useUID)
        

    def copy(self, message_set, new_mailbox, useUID=0):
        "Copies a message to a new location"
        message_set = ids2str(message_set)
        new_mailbox = new_mailbox.encode('imap4-utf-7')
        if useUID:
            typ, data = self.connection.uid('copy', message_set, new_mailbox)
        else:
            typ, data = self.connection.copy(message_set, new_mailbox)
        result_check((typ, data))
        return data

    def expunge(self):
        "Expunges deleted messages from the selected mailbox"
        typ, data = self.connection.expunge()
        result_check((typ, data))
        return data

    def append(self, mailbox, message, flags=(), date_time=None):
        "Appends a rfc822 message to the selected mailbox"
        mailbox = mailbox.encode('imap4-utf-7')
        if date_time is None:
#            date_time = strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime())
            date_time = gmtime()
        flagsStr = ','.join(flags)
        con = self.connection
        con.append(mailbox, flags, date_time, message)

    # Folder handling

    def create(self, mailbox):
        "Create a new mailbox"
        mailbox = mailbox.encode('imap4-utf-7')
        typ, data = self.connection.create(mailbox)
        result_check((typ, data))
        return data

    def subscribe(self, mailbox):
        "Create a new mailbox"
        mailbox = mailbox.encode('imap4-utf-7')
        typ, data = self.connection.subscribe(mailbox)
        result_check((typ, data))
        return data

    def rename(self, oldmailbox, newmailbox):
        "Create a new mailbox"
        oldmailbox = oldmailbox.encode('imap4-utf-7')
        newmailbox = newmailbox.encode('imap4-utf-7')
        typ, data = self.connection.rename(oldmailbox, newmailbox)
        result_check((typ, data))
        return data

    def delete(self, mailbox):
        "Deletes old mailbox"
        mailbox = mailbox.encode('imap4-utf-7')
        typ, data = self.connection.delete(mailbox)
        result_check((typ, data))
        return data

    def noop(self):
        "Does nothing"
        typ, data = self.connection.noop()
        result_check((typ, data))
        return data
        

    def recent(self):
        "Number of recent messages"
        typ, data = self.connection.recent()
        result_check((typ, data))
        return int(data[0])
        

    def logout(self):
        "Logs out from the server"
        self.connection.logout()
        self._change_state(LOGOUT)
        del self.connection


# a borg/singleton for pooling connections.
# Necessary as login can take a looong time
class ConnectionPool:

    __shared_state = {}

    def __init__(self):
        self.__dict__ = self.__shared_state

    def getConnection(self, server_uri, user, password):
        if not hasattr(self, 'connections'):
            self.connections = {}
        user_key = (server_uri, user, password)
        user_connections = self.connections.setdefault(user_key, [])
        # first try and return a connection from the pool
        for i in range(len(user_connections)-1, -1, -1):
            # a connection can be timed out, try noop to see if it's still open
            connection = user_connections.pop(i)
            try:
                connection.noop()
                return connection
            except:
                pass
        # if no usable connections in the pool, create new one and return it
        return Imap4Connector(server_uri, user, password)


    def release(self, connection):
        # take the connection and put it in the pool
        user_key = (connection.server_uri, connection.user, connection.password)
        user_connections = self.connections.setdefault(user_key, [])
        user_connections.append(connection)


    def logout_unused(self):
        if not hasattr(self, 'connections'):
            self.connections = {}
        # first try and return a connection from the pool
        for user_connections in self.connections.values():
            for i in range(len(user_connections)-1, -1, -1):
                # a connection can be timed out, try noop to see if it's still open
                connection = user_connections.pop(i)
                try:
                    connection.logout()
                except:
                    pass


#    def logout_unused(self):
#        "Release all unused connections"
#        for user_key, user_connections in self.connections.items():
#            for  in connections:
#                connection
#                try:
#                    connection.logout()
#                except:
#                    pass
#                del self.connections[user_key]


def Imap4Connector(server_uri, user, password):

    """
    A factory  that Connects to an imap server, and returns an imap client in 
    the connected state, or raises a 'IMAP4Exception' Exception.
    
    Normally it will return the base ImapClientConnection, but depending on the 
    server it connects to, it might return a special ImapConnection that can 
    handle server specific special cases.
    
    The interface will allways be the same for any client.
    """

    try:
        mailconn = imaplib.IMAP4(server_uri)
    except:
        raise IMAP4Exception, 'Cannot connect to mailserver'
    typ, data = mailconn.login(user, password)
    result_check((typ, data))
    connection = ImapClientConnection(mailconn)
    connection.setAuthData(server_uri, user, password)
    return connection



if __name__=='__main__':
    
    # credentials are imported from file not in SVN. So create your own to run tests
    from imaplib2credentials import servername, username, password

#    from email.Message import Message
#    msg = Message()
#    msg['From'] = 'maxm@mxm.dk'
#    msg['To'] = 'maxm@mxm.dk'
#    msg['Subject'] = 'Add to sent box'
#    msg.set_payload('This is the message: "42"')
#    
#    c = Imap4Connector(servername, username, password)
#    c.append(u'Sent', msg.as_string(), ('\\Seen',))
#    c.select(u'Sent')
#    c.logout()
    

##    c = Imap4Connector(servername, username, password)
##    c.select(readonly=1)
##    uids = [uid for (uid, flags, size) in c.getAllUids()]
###    results = c.fetch(ids2str(uids), '(BODYSTRUCTURE)', useUID=1)
###    results = c.fetchHeaders(uids, useUID=1)
##    for uid in uids:
##        results = c.fetchBodyPart(uid, 1, useUID=1)
##        for uid, content in results:
##            print repr(content)
##            print '\n'*5
###    for content in results:
###        print content
####        print parseNestedParens(content)
###        print ''
##    c.logout()

    

#    pool = ConnectionPool()
#    
#    time_connect = time()
#    c = pool.getConnection(servername, username, password)
#    print 'Authentication time:', time() - time_connect

#    time_select = time()
#    c.select('INBOX', readonly=1)
#    print 'selection', time() - time_select
#
#    print c.search(criteria=['UNSEEN'], useUID=1)

#    for folder in c.list():
#        time_select = time()
#        print 'Folder:', repr(folder.name)
#        print "n messages: %s" % (c.select(folder.name, readonly=1))
#        print 'selection time:', time() - time_select
#
#    c.logout()


#    uids = [uid for (uid, flags, size) in c.getAllUids()]
#    print uids
#    results = c.fetchRFC822(uids, useUID=1)
    #pool.release(c)
#    for uid, content in results:
#        print uid,

        


###    uids = [1719, 1740, 1694, 2050, 2051]
###    uids = [33, 34, 35, 36]
##    from time import time
##    c = Imap4Connector(servername, username, password)
###    c = Imap4Connector(servername, username, password)
##    start = time()
##    print c.select('Calendar', readonly=1)
##    c.logout()
##    print time() - start


#    c = Imap4Connector(servername, username, password)
#    c.select('Calendar', readonly=1)
#    uids = [uid for (uid, flags, size) in c.getAllUids()]
#    from test_uids import uids
#    uids = uids[:10]
#    results = c.fetchRFC822(uids, useUID=1)
#    results = c.fetchBody(uids, useUID=1)
#    part = c.fetchBodyPart(218, ('',), useUID=1)
#    print part
##    fields = ['From', 'To', 'Subject', 'In-Reply-To', 'Date']
##    results = c.fetchHeaders(uids, fields, useUID=1)
#    for uid, content in results:
#        print '########################################'
#        print content
#        print '----------------------------------------'
##
###    for msg in messages:
###        print msg
####    for uid, msg in messages:
####        print uid
####        print msg
##
###    for folder in c.list():
###        print '###############'
###        print repr(folder.name)
###        print c.select(folder.name)
###        print c.getAllUids()
####        print folder.flags
####        print folder.delimiter
####        print 'unreads:', 
####        print ''
##    c.logout()
##
##
#####    # test the calendar folder
#####    c = Imap4Connector(servername, username, password)
#####    c.select('Calendar')
#####    uids = [uid for (uid, flags, size) in c.getAllUids()]
#####    messages = c.fetchRFC822(uids, useUID=1)
#####    for uid, msg in messages:
#####        print '###########################################'
#####        print uid
#####        print msg
