# -*- coding: latin-1 -*-

"""
This is an Imap4 mail client for Plone.

It stores message headers as ordinary Python dicts. The message body is not 
stored in Zope. Whenever a message is viewed, it is fetched from the Imap 
server.

This approach has the following advantages:

    - It is as least as fast as the Zope server, as new message headers are only 
    fetched once in a while. This can be especially important if there is a slow 
    line between the Imap and the Zope server.
    
    Many mail users have several thousand messages in their mailboxes. It can be 
    a drag to fetch them all from the server if headers are not cached locally.
    
    - You can compare the cached message headers with the messages on the 
    server, making it possible to only fetch headers for changed/new messages. 
    Minimizing traffic between server and client.
    
    - Leaving the message body on the Imap server, means that attachments also 
    stays, so that you don't need to bloat the ZODB with thousand of multi-
    megabyte files.

    - Message subjects can be searched on the client
    
    - Making interface changes is simple for Plone developers, as they mostly 
    work with normal Plone folders and content objects. They also don't need to 
    know about special Imap formats.
    
    - It can make threaded views without fetching all message headers


It works like this:

    A mxmImapClient is the root folder. It *is* the Imap client. Every n 
    minute, it fetches new folders and message from the Imap server.
    
    mxmImapFolders are saved under the mxmImapClient folder, and 
    headers are stored as dicts in a IOBTree on the mxmImapFolder.
    
    When you view a message the mxmImapFolder returns a wrapper on a Python 
    email message. All values returned from the wrapper are in "Plone 
    compatible" formats. Eg. most are converted to the site encoding. Normally 
    utf-8. Making it easier to handle.

TODO:

    Sorting by
        title
        sender
        date
        thread (via 'In-Reply-To' header)

"""

# Python
from DateTime import DateTime
import time # for debugging only XXX
import email, smtplib
from email import Encoders
from email.Message import Message
from email.Header import Header
from email.MIMEMultipart import MIMEMultipart
from email.MIMEText import MIMEText
from cStringIO import StringIO
from sets import Set

# Zope
from AccessControl import ClassSecurityInfo
from BTrees.OOBTree import OOBTree

# CMF
from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.CMFCorePermissions import View, ListFolderContents
from Products.CMFCore.CMFCorePermissions import setDefaultRoles
from Products.CMFCore.utils import getToolByName

# Archetypes
from Products.Archetypes.public import BaseBTreeFolder, registerType, DisplayList

# Product
from schemas import message_client_schema
from config import PROJECTNAME, VIEW_CONTENTS_PERMISSION,\
                   EDIT_CONTENTS_PERMISSION, ADD_CONTENTS_PERMISSION
import imaplib2

##### testing and timing XXX remember to delete

ONE_MINUTE = 1.0/24/60
ANCIENT_TIME = (1970,1,1)

HEADER_FIELDS = ['From', 'To', 'Subject', 'In-Reply-To', 'Date']

def toHex(ust):
    "Converts unicode string to hex"
    return ust.encode('utf-8').encode('hex')

def fromHex(st):
    "Converts hex encoded utf-8 string to utf-8"
    return st.decode('hex').decode('utf-8')

def pathFromHex(path):
    "Converts a server path to hex"
    return [fromHex(id) for id in path]

def pathToHex(path):
    "Converts a server path to hex"
    return [toHex(id) for id in path]


class mxmImapClient(BaseBTreeFolder):

    content_icon = 'mxmImapClient_icon.gif'

    global_allow = 1

    schema = message_client_schema

    left_slots = (
        'here/mxmImapClient_portlet/macros/portlet',
    )

    allowed_content_types = ('mxmImapFolder','mxmImapMessage')

    actions = (
        {
            'id': 'view',
            'name': 'View',
            'action': 'string:${object_url}/mxmImapClient_view',
            'permissions': (View,)
        },
        {
            'id': 'metadata',
            'name': 'Properties',
            'action': 'string:${object_url}/base_metadata',
            'permissions': (EDIT_CONTENTS_PERMISSION,),
            'visible':0,
        },
    )

    # paths to hidden folders
    _hidden_folders = ()

    # Standard security settings
    security = ClassSecurityInfo()
    security.declareObjectProtected(View)


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getMxmMessageClient')
    def getMxmMessageClient(self):
        "helper for getting the client through aq"
        return self

    
    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'setPassword')
    def setPassword(self, password=''):
        if password:
            self.password = password


    def _timeToSync(self, context):
        "Returns true if it is time to sync"
        return 1
        # if forced sync, dont care about time
        if self.REQUEST.get('mxmForcedSync', 0):
            return 1
        # check if it is time to synch with server
        UPDATE_INTERVAL = ONE_MINUTE * self.getUpdateInterval()
        now = DateTime()
        ses = self.REQUEST.SESSION
        # there can be more than one messageClient per user
        session_key = 'mxm_last_updated_%s' % '/'.join(context.getPhysicalPath())
        last_updated = DateTime(*ses.get(session_key, ANCIENT_TIME))
        delta_time = now - last_updated
        do_sync = delta_time > UPDATE_INTERVAL
        if do_sync:
            ses[session_key] = now.parts()
        return do_sync


    def _addFolder(self, folder):
        "Add folders 'recursively' from path"
        f = self
        path = folder.path()
        for i in range(len(path)):
            id = path[i]
            if id: # some ids from Exchange are empty ??
                hexId = toHex(id)
                if not hasattr(f.aq_base, hexId):
                    f.invokeFactory('mxmImapFolder', hexId)
                f = getattr(f, hexId)
                f.setTitle(id)
                # set the Imap name of the folder
                f.folder_name = folder.delimiter.join(path[:i+1])
                f.folder_delimiter = folder.delimiter


    def _syncFolderList(self, connection):
        "Syncronizes the folder list"
        # get all client folders as relative to self.
        msgClientPath = list(self.getPhysicalPath())
        client_paths_hex = [list(f.getPhysicalPath()[len(msgClientPath):]) for f in self.getMessageFolders()]
        client_paths = [pathFromHex(client_path_hex) for client_path_hex in client_paths_hex]
        # figure out which folders to add to the client.
        serverFolders = connection.list()
        paths_to_add = [folder.path() for folder in serverFolders if folder.path() not in client_paths]
#        f = open('c:/temp/imap-folders.txt', 'w')
        for serverFolder in serverFolders:
#            print >> f, '========================'
#            print >> f, serverFolder
            if serverFolder.path() not in client_paths:
                self._addFolder(serverFolder)
#        f.close()
        # remove the folders that are deleted from the server. We do it with a 
        # backwards sorted list of paths so we dont accidently delete parent 
        # folders first
        server_paths_hex = [pathToHex(serverFolder.path()) for serverFolder in serverFolders]
        paths_to_remove = [path for path in client_paths_hex if path not in server_paths_hex]
        paths_to_remove.sort() # zope ids are in hex, so it must be sorted in hex
        paths_to_remove.reverse()
        for path in paths_to_remove:
            folder_id = path[-1]
            parentPath = '/'.join(msgClientPath + path[:-1])
            parent = self.unrestrictedTraverse(parentPath)
            parent.manage_delObjects([folder_id])
        get_transaction().commit()
        return serverFolders


    def _syncMessageHeaders(self, connection, serverFolders):
        "Syncronize the message headers in each folder"
        clientPath = list(self.getPhysicalPath())
        for folder in serverFolders:
            serverFolderPath = folder.name.split(folder.delimiter)
            clientFolderPath = '/'.join(clientPath + pathToHex(serverFolderPath))# ids are stored as hex
            clientFolder = self.unrestrictedTraverse(clientFolderPath)
            message_count = connection.select(folder.name, readonly=1)
            if message_count: # don't fetch uids if no messages
                infos = connection.getAllUids()
            else:
                infos = []
            
            # find server uids
            serverUidsMap = {}
            for info in infos:
                uid, flags, size = info
                # info[0] is the uid
                # messages flagged for deletion should not be seen on client
                if not r'\Deleted' in flags:
                    serverUidsMap[uid] = info
            serverUids = serverUidsMap.keys()
            
            # find client uids
            clientUidsMap = {}
            for clientId in clientFolder.getCacheUids():
                clientUidsMap[clientId] = 1
            clientUids = clientUidsMap.keys()
            # find out which messages to remove from the client folder
            idsToRemove = [uid for uid in clientUids if not serverUidsMap.get(uid, None)]
            
            # first delete from client
            clientFolder.delCaches(idsToRemove)

            # then add new messages
            # find out which messages to add to the client folder
            idsToAdd = [uid for uid in serverUids if not clientUidsMap.get(uid, None)]

#            # limit the number of messages that are fetched to 50 at a time
#            idsToAdd.sort()
#            idsToAdd = idsToAdd[-50:]
#            print 'idsToAdd', idsToAdd

            # then fetch the headers and add them :-)
            if idsToAdd:
                messages = connection.fetchHeaders(idsToAdd, HEADER_FIELDS, useUID=1)
                n_added = 0
                messageItems = []
                clientFolder.setCachesFromMessages(messages)
                # set unread flags
                for key, (uid, flags, size) in serverUidsMap.items():
                    if not '\\Seen' in flags:
                        self.markAsUnread(clientFolder, uid)
                # do a partial commit for each folder
                get_transaction().commit()
    

    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'hideSelectedFolders')
    def hideSelectedFolders(self, folderPaths=None, REQUEST=None):
        "Hide selected folders via path"
        if folderPaths is None:
            folderPaths = []
        self._hidden_folders = folderPaths
        if not REQUEST is None:
            self.REQUEST.RESPONSE.redirect(self.absolute_url() + '/mxmImapClient_hideFolders')
        

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getHiddenFolders')
    def getHiddenFolders(self):
        "Returns the hidden folders"
        return self._hidden_folders


    def getUpdatedMessagesUids(self, folder):
        "Gets uids of new and deleted messages in a folder"
        connection = self.getConnection()
        # get header info
        message_count = connection.select(folder.getFolderName(), readonly=1)
        if message_count: # don't fetch uids if no messages
            infos = connection.getAllUids()
        else:
            infos = []
        # find server uids
        serverUidsMap = {}
        for info in infos:
            uid, flags, size = info
            # info[0] is the uid
            # messages flagged for deletion should not be seen on client
            if not r'\Deleted' in flags:
                serverUidsMap[uid] = info
        serverUids = serverUidsMap.keys()
        # find client uids
        clientUidsMap = {}
        for clientId in folder.getCacheUids():
            clientUidsMap[clientId] = 1
        clientUids = clientUidsMap.keys()
        # find out which messages to remove from the client folder
        idsToRemove = [uid for uid in clientUids if not serverUidsMap.get(uid, None)]
        # find out which messages to add to the client folder
        idsToAdd = [uid for uid in serverUids if not clientUidsMap.get(uid, None)]
        self.releaseConnection(connection)
        return (idsToAdd, idsToRemove)


    def getConnection(self):
        "Returns a connection to the server."
        server = self.getIncommingServer()
        user = self.getUsername()
        password = self.getPassword()
        pool = imaplib2.ConnectionPool()
        connection = pool.getConnection(server, user, password)
        return connection


    def releaseConnection(self, connection):
        "Releases a connection"
        pool = imaplib2.ConnectionPool()
        pool.release(connection)


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'sync')
    def sync(self, context=None):
#        return # for debug
        """
        Syncronizes the messageclient with the remote server. Call this once per 
        page-load, and the client should keep in sync. It doesn't hurt the 
        clients speed to call it more than once. But it's unnecessary.
        """
        # context should only be messageFolder or Client
        if context is None:
            syncFolder = self
        else:
            syncFolder = getattr(context, 'getMxmMessageFolder', self.getMxmMessageClient)()
        if self._timeToSync(syncFolder):
            server = self.getIncommingServer()
            user = self.getUsername()
            password = self.getPassword()
            if server and user and password:
#                try:
                connection = self.getConnection()
                connect_ok = 1
#                except:
#                    connect_ok = 0
#                    print 'Wrong data, server: %s, user: %s, password: %s' % (server, user, password)
                if connect_ok:
                    # make shure that the folder list on the client == server
                    serverFolders = self._syncFolderList(connection)
                    # synchronize the message headers in serverFolders
                    if context is None: # sync all folders!
                        foldersToSync = serverFolders
                    else: # just sync the one folder passed as context
                        base = syncFolder.aq_base
                        foldersToSync = []
                        if hasattr(base, 'portal_type'):
                            if base.portal_type == 'mxmImapClient':
                                foldersToSync = [f for f in serverFolders if f.name.upper() == 'INBOX']
                            elif base.portal_type == 'mxmImapFolder':
                                folder_name = base.getFolderName()
                                foldersToSync = [f for f in serverFolders if f.name == folder_name]                    
                    self._syncMessageHeaders(connection, foldersToSync)
                    self.releaseConnection(connection)


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'syncHeadersFromUids')
    def syncHeadersFromUids(self, folder, uids):
        """
        Gets a list of headers from a list of uids, and update them.
        """
        # then fetch the headers and add them :-)
        connection = self.getConnection()
        messages = connection.fetchHeaders(uids, HEADER_FIELDS, useUID=1)
        folder.setCachesFromMessages(messages)
        get_transaction().commit()
        self.releaseConnection(connection)


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'syncFolders')
    def syncFolders(self, REQUEST=None):
        "Syncronize the list of folders"
        connection = self.getConnection()
        self._syncFolderList(connection)
        self.releaseConnection(connection)
        if not REQUEST is None:
            self.REQUEST.RESPONSE.redirect(self.absolute_url())
        


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'fetchUnread')
    def fetchUnread(self, mxmFetchUnread=0, REQUEST=None):
        "Fetches unread headers to the inbox"
        message = 'No new unread messages'
        if mxmFetchUnread:
            # first sync folders
            self.syncFolders()
            # find inbox
            inbox = None
            folders = self.objectValues('mxmImapFolder')
            for folder in folders:
                if folder.getFolderName().lower() == 'inbox':
                    inbox = folder
                    break
            if not inbox is None:
                # find recent uids
                connection = self.getConnection()
                connection.select('INBOX', readonly=1)
                uids = connection.search(criteria=['UNSEEN'], useUID=1)
                self.setUnreads(inbox, uids)
                if uids:
                    idsToAdd = list(Set(uids) - Set(inbox.getCacheUids()))
                    messages = connection.fetchHeaders(idsToAdd, HEADER_FIELDS, useUID=1)
                    messageItems = []
                    inbox.setCachesFromMessages(messages)
                    # set unread flags
                    for uid in idsToAdd:
                        self.markAsUnread(inbox, uid)
                    # do a partial commit
                    get_transaction().commit()
                    message = '%s new unread message(s) found' % len(idsToAdd)
            self.releaseConnection(connection)
        if not REQUEST is None:
            self.REQUEST.RESPONSE.redirect(self.absolute_url() + '?portal_status_message=%s' % message)


    ##########################
    # methods for checking if messages are read

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getMessagePath')
    def getMessagePath(self, folder, uid):
        """
        Returns the message path relative to self. Relative paths makes it 
        possible to move a messageClient without loosing the paths to messages
        """
        return (self.getFolderPathJoined(folder), uid)


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getFolderPath')
    def getFolderPath(self, folder):
        """
        Returns the folder path relative to self. Relative paths makes it 
        possible to move a messageClient without loosing the paths to folder
        """
        return tuple(folder.getPhysicalPath()[len(self.getPhysicalPath()):])


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getFolderPathJoined')
    def getFolderPathJoined(self, folder):
        """
        Returns the joined folder path relative to self. Relative paths makes it 
        possible to move a messageClient without loosing the paths to folder
        """
        return '/'.join(self.getFolderPath(folder))


    def _getUnreadsDict(self):
        "Returns a dict with paths to unread messages as keys and true as value"
        if not hasattr(self, '_unreadMessages'):
            self._unreadMessages = OOBTree()
        return self._unreadMessages


    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'markAllAsRead')
    def markAllAsRead(self, REQUEST=None):
        "Mark all as read"
        self._unreadMessages = OOBTree()
        if REQUEST:
            message = 'All messages marked as read'
            REQUEST.RESPONSE.redirect(self.absolute_url() + '?portal_status_message=%s' % message)


    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'markAsRead')
    def markAsRead(self, folder, uid):
        "Sets a messages state as read"
        unreadsDict = self._getUnreadsDict()
        msgPath = self.getMessagePath(folder, uid)
        if unreadsDict.has_key(msgPath):
            del unreadsDict[msgPath]


    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'markAsUnread')
    def markAsUnread(self, folder, uid):
        "Sets a messages state as read"
        uid = int(uid)
        self._getUnreadsDict()[self.getMessagePath(folder, uid)] = 1


    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'setUnreads')
    def setUnreads(self, folder, uids):
        "Sets the unread messages for folder"
        newUids = Set(uids)
        oldUids = Set([id for (path, id) in self._getUnreadsDict().keys()])
        markAsUnRead = newUids - oldUids
        markAsRead = oldUids - newUids
        unreadsDict = self._getUnreadsDict()
        # mark as unread
        for uid in markAsUnRead:
            unreadsDict[self.getMessagePath(folder, uid)] = 1
        # mark as read
        for uid in markAsRead:
            msgPath = self.getMessagePath(folder, uid)
            if unreadsDict.has_key(msgPath):
                del unreadsDict[msgPath]


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'isUnread')
    def isUnread(self, folder, uid):
        "Returns true if message is unread"
        unreadsDict = self._getUnreadsDict()
        return unreadsDict.get(self.getMessagePath(folder, uid), 0)

    # getting unreads

#    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getMessageFromPath')
#    def getMessageFromPath(self, path, uid):
#        "Returns a local message from a relative messagepath"
#        folder = self.unrestrictedTraverse(path)
#        return folder.getCache(uid)


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getUnreadMessages')
    def getUnreadMessages(self):
        "Returns a list of unread message [(path, cache)]"
        unreads = []
        paths = self._getUnreadsDict().keys()
        uniq = {} # {path:[uids]}
        for path, uid in paths:
            uid = int(uid) # safetybelt
            uniq.setdefault(path, []).append(uid)
        for path, uids in uniq.items():
            folder = self.unrestrictedTraverse(path)
            unreads += [(path, cache) for cache in folder.getCaches(uids)]
        return unreads


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getUnreadMessagesSorted')
    def getUnreadMessagesSorted(self, reverse=0):
        "Returns a sorted list of unread message caches as a list of caches"
        messages = self.getUnreadMessages()
        decorated = [(message['Date'], path, message) for (path, message) in messages]
        decorated.sort()
        sorted = [s[-2:] for s in decorated]
        if reverse:
            sorted.reverse()
        return sorted


    ##############


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getMessageFolders')
    def getMessageFolders(self, depth=0, showHidden=0):
        "Recursive folder search. Special folders are allways first."
        # get and sort folders
        folders = self.objectValues('mxmImapFolder')
        decorated = [(f.title_or_id().lower(), f) for f in folders]
        decorated.sort()
        sorted = [d[-1] for d in decorated]
        # Sort into special and other folder, and get the subfolders
        specialFolderNames = self.getSpecialFolderNames()
        specialFolders = []
        otherFolders = []
        for folder in sorted:
            if showHidden or not folder.isHidden():
                if folder.getFolderName() in specialFolderNames:
                    specialFolders.append(folder)
                    specialFolders += folder.getMessageFolders(depth=depth+1, showHidden=showHidden)
                else:
                    otherFolders.append(folder)
                    otherFolders += folder.getMessageFolders(depth=depth+1, showHidden=showHidden)
        return specialFolders + otherFolders

    ##

    def _specialFolderName(self, folder, default):
        "Returns name of special folder, or default value"
        if folder and hasattr(folder, 'getFolderName'):
            return folder.getFolderName()
        return default

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getInboxName')
    def getInboxName(self):
        "Returns the name of the inbox"
        return self._specialFolderName(self.getInboxFolder(), 'INBOX')

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getSentboxName')
    def getSentboxName(self):
        "Returns the name of the sentbox"
        return self._specialFolderName(self.getSentboxFolder(), 'Sent')

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getTrashboxName')
    def getTrashboxName(self):
        "Returns the name of the sentbox"
        return self._specialFolderName(self.getTrashboxFolder(), 'Trash')

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getSpecialFolderNames')
    def getSpecialFolderNames(self):
        "Returns the name of the special folders"
        specialFolderNames = []
        a = specialFolderNames.append
        a(self.getInboxName())
        a(self.getSentboxName())
        a(self.getTrashboxName())
        return specialFolderNames

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getFolderDisplayList')
    def getFolderDisplayList(self):
        "Returns a displaylist for folders"
        folders = self.getMessageFolders()
        uids = [(f.UID(), f.getFolderName()) for f in folders]
        return DisplayList([('','Default')] + uids)

    ##

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'calculateDepth')
    def calculateDepth(self, obj):
        """
        Calculates the objects level below the messageclient. Very nifty for 
        that tree-like navigation look.
        """
        obj_depth = len(obj.getPhysicalPath())
        msgClientDepth = len(self.getPhysicalPath())
        depth = len(obj.getPhysicalPath()) - len(self.getPhysicalPath()) - 1
        if depth < 0:
            depth = 0
        return depth


    ################################################
    # Methods for manipulating folders on the server

    def _fname(self, path):
        # returns folder name from a path in unicode
        encoding = self._site_encoding()
        parent_folder = self.unrestrictedTraverse(path)
        return parent_folder.getFolderName()


    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'addFolder')
    def addFolder(self, parent_folder_path='', new_folder_name='', REQUEST=None):
        "Add a new folder under parent folder"
        if new_folder_name:
            # add as subfolder to parent folder
            encoding = self._site_encoding()
            if parent_folder_path:
                parent_folder = self.unrestrictedTraverse(parent_folder_path)
                parent_folder_name = parent_folder.getFolderName().encode(encoding)
                delimiter = parent_folder.getFolderDelimiter()
                new_folder_name = parent_folder_name + delimiter + new_folder_name
            uName = new_folder_name.decode(encoding)
            try:
                connection = self.getConnection()
                connection.create(uName)
                connection.subscribe(uName)
                self._syncFolderList(connection)
                self.releaseConnection(connection)
                message = 'Folder "%s" added' % (new_folder_name)
            except:
                message = 'ERROR: The folder "%s" could not be added' % (new_folder_name)
        # make error message    
        else:
            message = 'ERROR: You must give the new folder a name'
        # return the zope way
        if REQUEST:
            REQUEST.RESPONSE.redirect(self.absolute_url() + '?portal_status_message=%s' % message)


    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'renameFolder')
    def renameFolder(self, parent_folder_path=None, new_folder_name='', REQUEST=None):
        "Rename a folder"
        if new_folder_name:
            parent_folder = self.unrestrictedTraverse(parent_folder_path)
            parent_folder_name = parent_folder.getFolderName()
            encoding = self._site_encoding()
            connection = self.getConnection()
            connection.rename(parent_folder_name, new_folder_name.decode(encoding))
            self._syncFolderList(connection)
            self.releaseConnection(connection)
            message = 'Folder "%s" renamed to "%s"' % (parent_folder_name.encode(encoding), new_folder_name)
        elif not parent_folder_path:
            message = 'ERROR: No folder selected to rename'
        else:
            message = 'ERROR: Folder could not be renamed to "%s"' % new_folder_name
        if REQUEST:
            REQUEST.RESPONSE.redirect(self.absolute_url() + '?portal_status_message=%s' % message)


    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'deleteFolder')
    def deleteFolder(self, parent_folder_path=None, REQUEST=None):
        "Add a new folder under parent folder"
        if parent_folder_path == '/':
            message = 'ERROR: Cannot delete the root folder'
        else:
            encoding = self._site_encoding()
            parent_folder = self.unrestrictedTraverse(parent_folder_path)
            parent_folder_name = parent_folder.getFolderName()
            if parent_folder_name in self.getSpecialFolderNames():
                message = 'ERROR: %s is a special folder and cannot be deleted' % parent_folder_name
            else:
                try:
                    # delete on the server
                    connection = self.getConnection()
                    connection.delete(parent_folder_name)
                    self._syncFolderList(connection)
                    self.releaseConnection(connection)
                    message = 'Folder "%s" deleted' % parent_folder_name.encode(encoding)
                except:
                    message = 'ERROR: It was not possible to delete "%s"' % parent_folder_name.encode(encoding)
        if REQUEST:
            REQUEST.RESPONSE.redirect(self.absolute_url() + '?portal_status_message=%s' % message)


    #################################################
    # Methods for manipulating messages on the server

    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'deleteMessages')
    def deleteMessages(self, folder, ids):
        """
        if current folder is trashcan, Expunge the messages.
        else move messages to trashcan
        """
        connection = self.getConnection()
        connection.select(folder)
        trashboxName = self.getTrashboxName()
        if folder != trashboxName:
            connection.copy(ids, trashboxName, useUID=1)
        connection.storeAddFlags(ids, ('\Deleted',), silent=1, useUID=1)
        connection.expunge()
        self.releaseConnection(connection)


    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'moveMessages')
    def moveMessages(self, from_folder, ids, to_folder):
        """
        Moves messages from one folder to another
        """
        connection = self.getConnection()
        connection.select(from_folder)
        connection.copy(ids, to_folder, useUID=1)
        connection.storeAddFlags(ids, ('\Deleted',), silent=1, useUID=1)
        connection.expunge()
        self.releaseConnection(connection)


    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'moveMessage')
    def moveMessage(self, from_folder, id, to_folder, REQUEST=None):
        """
        Move a single message and return to from folder. Called from the message view.
        """
        if not REQUEST is None:
            REQUEST.RESPONSE.redirect('%s/%s' % (self.absolute_url(), from_folder))


#    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'fetchMessage')
#    def fetchMessage(self, msg):
#        """
#        Takes an mxmImapMessage and fetches a message from the server and returns it 
#        as an email message.
#        """
#        msgId = msg.getId()
#        folderName = msg.getFolderName()
#        connection = self.getConnection()
#        connection.select(folderName)
#        msg = connection.fetchRFC822([msgId], useUID=1)
#        msg = msg[0][-1] # fetch returns [(uid, msg)]
#        self.releaseConnection(connection)
#        return email.message_from_string(msg)


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'fetchMessage')
    def fetchMessage(self, folderName, uid):
        """
        Takes an mxmImapMessage and fetches a message from the server and returns it 
        as an email message.
        """
        connection = self.getConnection()
        connection.select(folderName)
        msg = connection.fetchRFC822([uid], useUID=1)
        msg = msg[0][-1] # fetch returns [(uid, msg)]
        self.releaseConnection(connection)
        return email.message_from_string(msg)


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'fetchMessage')
    def fetchMessageText(self, folderName, uid):
        """
        Takes an mxmImapMessage and fetches a message from the server and returns it 
        as an email message.
        """
        connection = self.getConnection()
        connection.select(folderName)
        msg = connection.fetchRFC822Text([uid], useUID=1)
        msg = msg[0][-1] # fetch returns [(uid, msg)]
        self.releaseConnection(connection)
        return email.message_from_string(msg)


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'fetchMessageFromPath')
    def fetchMessageFromPath(self, folderPath, uid):
        """
        Takes a message path and fetches a message from the server and returns it 
        as an email message.
        """
        folder = self.unrestrictedTraverse(folderPath)
        folderName = folder.getFolderName()
        connection = self.getConnection()
        connection.select(folderName)
        msg = connection.fetchRFC822([uid], useUID=1)
        msg = msg[0][-1] # fetch returns [(uid, msg)]
        self.releaseConnection(connection)
        return email.message_from_string(msg)

    def _headerToDefaultCharset(self, header):
        "Returns an encoded header"
        encoding = self._site_encoding()
        try:
            parts = email.Header.decode_header(header)
            new_header = email.Header.make_header(parts)
            return unicode(new_header).encode(encoding)
        except:
            try:
                return tryDecoding(header, ['utf-8', 'latin-1']).encode(encoding)
            except:
                return 'ENCODING ERRORS!'

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getAttachment')
    def getAttachment(self, folderPath, uid, number):
        "Returns attachment with number from message"
        # ok this is a hack, but most messages has only one large attachment
        # So most likely no one will ever notice.
        # (I'm excusing this with "no premature optimization")
        RESPONSE = self.REQUEST.RESPONSE 
        message = self.fetchMessageFromPath(folderPath, uid)
        maintype = message.get_type()
        if maintype == 'application/octet-stream' :
            RESPONSE.setHeader('Content-Type', maintype)
            RESPONSE.setHeader('Content-Length', str(message.get_size))
            filename = message.get_param('name', "attachment %s" % number)
            filename = self._headerToDefaultCharset(filename)
            RESPONSE.setHeader('Content-Disposition', 'filename=%s' % filename)
            return message.get_payload(decode=1)
        counter = 0
        for part in message.walk():
            counter += 1
            if counter == number:
                part_type = part.get_type()
                if part_type == 'message/rfc822' :
                    part_type = 'text/plain'
                RESPONSE.setHeader('Content-Type', part_type)
                filename = part.get_param('name', "attachment %s" % number)
                #filename = 'attachment'
                RESPONSE.setHeader('Content-Disposition', 'filename=%s' % filename)
                return part.get_payload(decode=1)


    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'appendToSentBox')
    def appendToSentBox(self, msg):
        "Appends a message to the Sent box"
        Sent = self.getSentboxName()
        connection = self.getConnection()
        try:
            connection.append(Sent, msg.as_string(), ('\\Seen',))
        except:
            # we don't want to not send messages just because the 'Sent' box isn't set correctly
            pass
        self.releaseConnection(connection)


    def _site_encoding(self):
        "Returns the site encoding"
        portal_properties = getToolByName(self, 'portal_properties')    
        site_props = portal_properties.site_properties
        return site_props.default_charset or 'utf-8'

    def _mailhost(self):
        # tries to fetch a mailhost/smtp server url from several locations
        # try user setting
        mailhost = self.getSmtpServer()
        if mailhost:
            return mailhost
        # try site setting
        if hasattr(self, 'MailHost'):
            mh = self.MailHost
            mailhost = mh.smtp_host
            if mailhost:
                return mailhost
        # desperately try sending via localhost
        return 'localhost'

    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'sendMessage')
    def sendMessage(self, to=None, cc=None, bcc=None, inReplyTo=None, 
                    subject=None, body='', attachments=None, mode=None, 
                    folderPath=None, uid=None, REQUEST=None):
        "Sends a message"
        site_encoding = self._site_encoding()
        ##################
        # Create the message
        msg = Message()
        msg.set_payload(body, site_encoding)
        #####################################
        # if attachment, convert to multipart
        # file fields are posted even if empty, so we need to remove those :-s
        if attachments is None:
            attachments = []
        attachments = [a for a in attachments if a] 
        if attachments or (mode == 'Forward'):
            mimeMsg = MIMEMultipart()
            mimeMsg.attach(msg)
            # If mode is Forward we get original message as attachment
            if mode == 'Forward' and folderPath and uid:
                original = self.fetchMessageFromPath(folderPath, uid)
                mimeMsg.attach(original)
            for attachment in attachments:
                # Add the attachment
                tmp = email.message_from_string(str(attachment.headers))
                filename = tmp.get_param('filename', 'Attachment', 'Content-Disposition')
                # clean up IE paths
                filename = filename[max(filename.rfind('/'),
                                        filename.rfind('\\'),
                                        filename.rfind(':')
                                       )+1:]
                contentType = tmp['Content-Type']
                attach_part = Message()
                attach_part.add_header('Content-Type', contentType, name=filename)
                attach_part.add_header('Content-Disposition', 'attachment', filename=filename)                
                attach_part.set_payload(attachment.read())
                Encoders.encode_base64(attach_part)
                mimeMsg.attach(attach_part)
            msg = mimeMsg
        ########################
        # set headers on message
        from_ = self.getEmail()
        msg['From'] = from_
        msg['Reply-To'] = from_
        if to: msg['To'] = to
        if cc: msg['Cc'] = cc
        if bcc: msg['Bcc'] = bcc
        msg['Date'] = self.ZopeTime().rfc822() # needed by some servers
        if inReplyTo:
            msg['In-Reply-To'] = inReplyTo
        msg['Subject'] = Header(subject, site_encoding)
        ##################
        # Send the message
        SMTPserver = self._mailhost()
        outbox = smtplib.SMTP(SMTPserver)
        try:
            cleaner = lambda adresses: [adress.strip() for adress in adresses.split(',') if adress.strip()]
            all_receivers = cleaner(to) + cleaner(cc) + cleaner(bcc)
            outbox.sendmail(from_, all_receivers, msg.as_string())
            success = 1
        except:
            success = 0
        ######################################
        # append the message to the 'Sent' box
        if success:
            Sent = self.getSentboxName()
            connection = self.getConnection()
            try:
                connection.append(Sent, msg.as_string(), ('\\Seen',))
            except:
                # we don't want to not send messages just because the sentbox isn't set correctly
                pass
        #################################
        # returns a portal status message
        if REQUEST:
            if success:
                message = 'Succes! The message was sent '
            else:
                message = 'Error! The message could not be sent'
            REQUEST.RESPONSE.redirect(self.absolute_url() + '/mxmImapClient_compose?portal_status_message=%s' % message)


    """
    Helper methods for the composer. These methods returns the values for the 
    fields in the compose windows depending on the mode. The values from the 
    emails are passed through the form, to avoid a trip to the message server.
    """

    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'composerMessagePath')
    def composerFolderPath(self):
        "Returns the value for the text field"
        return self.REQUEST.get('folderPath', '')

    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'composerMode')
    def composerMode(self):
        "Returns the composer mode"        
        req = self.REQUEST
        if req.has_key('mode_Reply'):
            return 'Reply'
        if req.has_key('mode_Reply_all'):
            return 'Reply all'
        if req.has_key('mode_Forward'):
            return 'Forward'
        return 'Compose'
        
    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'composerTo')
    def composerTo(self):
        "Returns the value for the text field"
        mode = self.composerMode()
        req = self.REQUEST
        NewTo = []
        if mode in ['Reply', 'Reply all']:
            From = req.get('From', '')
            if From:
                NewTo.append(From)
        if mode == 'Reply all':
            To = req.get('To', '')
            if To:
                NewTo.append(To)
        return ', '.join(NewTo)

    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'composerCc')
    def composerUid(self):
        "Returns the value for the text field"
        req = self.REQUEST
        mode = self.composerMode()
        if mode in ['Reply', 'Reply all', 'Forward']:
            return req.get('uid', '')
        return ''

    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'composerCc')
    def composerCc(self):
        "Returns the value for the text field"
        req = self.REQUEST
        mode = self.composerMode()
        if mode == 'Reply all':
            return req.get('Cc', '')
        return ''

    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'composerText')
    def composerText(self):
        "Returns the value for the text field"
        mode = self.composerMode()
        if mode in ['Reply', 'Reply all']:
            return self.REQUEST.get('wrappedText', '')
        return ''

    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'composerInReplyTo')
    def composerInReplyTo(self):
        "Returns the value for the text field"
        mode = self.composerMode()
        if mode in ['Reply', 'Reply all']:
            return self.REQUEST.get('messageId', '')
        return ''

    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'composerSubject')
    def composerSubject(self):
        "Returns the value for the Reply to header value"
        mode = self.composerMode()
        if mode in ['Reply', 'Reply all']:
            return 'Re: ' + self.REQUEST.get('Subject', '')
        if mode == 'Forward':
            return 'Fwd: ' + self.REQUEST.get('Subject', '')
        return ''


    # mxmContacts and mxmImapClient should really be more integrated !!!

    def _contactPersons(self):
        # returns dict of contact persons {email:full name}
        # get fromn catalog
        catalog = getToolByName(self, 'portal_catalog')
        brains = catalog(portal_type='mxmContactsPerson')
        objects = []
        for brain in brains:
            try:
                obj = brain.getObject()
                if not obj is None:
                    objects.append(obj)
            except:
                pass
        # Put into dict
        contacts = {}
        for object in objects:
            contacts[object.getEmail()] = object.Title().strip()
        return contacts



    def composerContactsOptions(self):
        "Renders a select widget for contact emails. Its cached"
        contactPersons = self._contactPersons()
        if not contactPersons:
            return ''
        out = StringIO()
        keys = contactPersons.keys()
        keys.sort()
        for email in keys:
            if email:
                name = contactPersons[email]
                if name:
                    item = '%s &lt;%s&gt;' % (name, email)
                else:
                    item = email
                out.write('<option value="%s">%s</option>' % (email, item))
        return out.getvalue()



setDefaultRoles( ADD_CONTENTS_PERMISSION, ('Manager', 'Owner',))
setDefaultRoles(EDIT_CONTENTS_PERMISSION, ('Manager', 'Owner',))
setDefaultRoles(VIEW_CONTENTS_PERMISSION, ('Manager', 'Owner',))

registerType(mxmImapClient, PROJECTNAME)

